// Copyright - Simovits Consulting 2009
// Author: James Dickson.

#include <winsock2.h>
#include <direct.h>
#include <string>
#include <openssl/err.h>

#include "common.h"
#include "sslserver.h"

// Globals
char *pass;

/** 
 * Listen function.
 */
SOCKET startListen(int port)
{
    SOCKET sock;
    struct sockaddr_in sin;
	sockaddr_in service;
    
    
	// Create socket
    if((sock = socket(AF_INET, SOCK_STREAM, 0))<0)
	{
		return -1;
	}
    
	// Setup socket
    memset(&sin, 0, sizeof(sin));	
	service.sin_family = AF_INET; 
	service.sin_addr.s_addr = INADDR_ANY; 
	service.sin_port = htons(port);

	// Bind 
	if (bind(sock, (SOCKADDR*)&service, sizeof(service)) == SOCKET_ERROR)
	{
		printf("[-] ERRROR: bind() failed: %ld.\n", WSAGetLastError());
		closesocket(sock);
		return -1;
	}

	// Start listen
	if (listen( sock , 1) == SOCKET_ERROR)
	{
		printf("[-] ERROR: listen(): Error listening on socket %ld.\n", WSAGetLastError());
	}
	else
	{
		printf("[+] Waiting for connections on port %d. To exit press <Ctrl+C> ...\n", port);
	} 

    return sock ;
}

/**
 * Serve the HTTP-page.
 */
static int pushHTML(SSL *ssl, SOCKET s, char *strCaller, int bits, char *strCipher, char *strProtocol, TESTINFO testInfo)
{
	const int BSIZE = 4096;
	char buf[BSIZE];
	int retval = 0;
	BIO *io;
	BIO *ssl_bio;

	io=BIO_new(BIO_f_buffer());
	ssl_bio=BIO_new(BIO_f_ssl());
	BIO_set_ssl(ssl_bio, ssl, BIO_CLOSE);
	BIO_push(io, ssl_bio);

	// Wait for request...
	while(TRUE)
	{
		retval = BIO_gets(io, buf, BSIZE-1);

		int sslError = SSL_get_error(ssl, retval);

		if(sslError != SSL_ERROR_NONE)
		{
			printf("[-] ERROR: SSL read error %d \n", sslError);
			break;
		}

		if(testInfo.bVerbose)
		{
			printf("[+] Recv: %s \n", buf);
		}


		// Find the end of HTTP headers 
		if(!strcmp(buf,"\r\n") || !strcmp(buf,"\n"))
		{
			break;
		}
	}

	// Write HTTP response.
	if((retval = BIO_puts(io, "HTTP/1.0 200 OK\r\n"))<=0)
	{
		printf("[-] ERROR: Write error. Could not write header.\n");
	}

	if((retval = BIO_puts(io,"Server: SimovitsSSLDiagnos \r\n\r\n"))<=0)
	{
		printf("[-] ERROR: Write error.\n");
	}

	// Web page to display.
	char strWebPage[4096];
	sprintf(strWebPage,
		"<html><body><h1>SSLDiagnos - SSL Server test.</h1> "
		"<br><b>If you see this then your client supports the server ciphers.</b> <br><br> "
		"- Client adress %s <br>"
		"- Cipher %s <br>"
		"- Bitcount %d <br>"
		"- Protocol %s <br>"
		"<br><hr>"
		"<i>Simovits Consulting </i> <a href=http://www.simovits.com>www.simovits.com</a>.</body></html>",
		strCaller,
		strCipher,
		bits,
		strProtocol
		);

	if((retval = BIO_puts(io, strWebPage)) <= 0)
	{
		printf("[-] ERROR: Write error. Could not write page\n");
	}

	if((retval = BIO_flush(io)) < 0)
	{
		printf("[-] ERROR: Error while flushing BIO\n");
	}



	return 0;
}


/**
 * Password callback
 */
static int password_cb(char *buf,int num, int rwflag,void *userdata)
{
	if(num < strlen(pass)+1) 
	{
		return 0;
	}

	strcpy(buf, pass);
	return strlen(pass);
}
/**
 * Verify certificate callback routine.
 */
static int cert_verify_callback_server(int ok, X509_STORE_CTX *ctx)
{
	printf("[+] Verifying cert...\n");
	return ok;
}


/**
 * This functionality can also be accomplished using the openssl command line tool.
 * However, we want to see exactly what happens thus this feature was added.
 */
void serverAcceptLoop(TESTINFO testInfo, RequestProtocol protocol)
{
    SOCKET sock,s;
    BIO *sbio;
    SSL_CTX *ctx;
    SSL *ssl;
    int r;    
    
    // Build our SSL context
    ctx=initCTX(protocol, testInfo.bUseWorkarounds, testInfo.bUseCiscoSpeshul);

	// Set context
	int sessionIdContext = 1;
    SSL_CTX_set_session_id_context(ctx, (unsigned char*)&sessionIdContext, sizeof sessionIdContext); 

    // Set our cipher list 
	SSL_CTX_set_cipher_list(ctx, testInfo.strServerCipher);
	SSL_CTX_set_options (ctx,SSL_OP_ALL);

	// Set the keyfile and password.
	char*keyfile = testInfo.strServerCertPath;
	char*privatekeyfile = testInfo.strServerCertKeyfile;
	char *password = testInfo.strServerCertPassword;
	pass=password;

    // Load our keys and certificates
    if(!(SSL_CTX_use_certificate_chain_file(ctx, keyfile)))
	{
		printf("[-] Error: Cannot load certificate file: %s\n", keyfile);
		return;
	}
   	
	// Set password
    SSL_CTX_set_default_passwd_cb(ctx, password_cb);

	// Never verify...
	SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, cert_verify_callback_server);
	

	// Load the private key
    if(!(SSL_CTX_use_PrivateKey_file(ctx, privatekeyfile, SSL_FILETYPE_PEM)))
	{
		printf("[-] ERROR: can not read key file: %s\n", privatekeyfile);
		return;
	}	  

    // Load the CAs we trust
    if(!(SSL_CTX_load_verify_locations(ctx, testInfo.strCACertPath, NULL)))
	{
		printf("[-] Error - Cannot find proper CA list file: %s", testInfo.strCACertPath);
		return ;
	}

   
	// Start to listen.
    sock = startListen(testInfo.port);
	sockaddr addrIncoming;
	int addrLen = sizeof(addrIncoming);

	// Accept loop...
    while(TRUE)
	{
		if((s=accept(sock, &addrIncoming, &addrLen)) < 0)
		{       
			return;
		}
		else 
		{	
			int bits = 0;
			char strName[256];
			char * strShortname;
			char *strProtocol;

			ssl=SSL_new(ctx);
			SSL_set_fd(ssl, s);
			SSL_set_accept_state(ssl);	// We work in server mode
			r=SSL_accept(ssl);			// Accept the connection.


			SSL_CIPHER *sslCipher = (SSL_CIPHER *) SSL_get_current_cipher(ssl);	// Fetch the current cipher
			

			if(sslCipher != NULL)
			{				
				SSL_CIPHER_description(sslCipher, strName, 255);
				strShortname = (char *) SSL_CIPHER_get_name(sslCipher);				
				bits = SSL_CIPHER_get_bits(sslCipher, &bits);

				printf("[+] Cipher used: %s\n", strShortname);
				printf("[+] Cipher bit count used: %d\n", bits);

				//SSL_METHOD *method = SSL_get_ssl_method(ssl);	

				strProtocol = (char *) SSL_get_version(ssl);

				printf("[+] Protocol used: %s\n", strProtocol);
			}

			// Get IP-info.
			char strInfo[1024];
			sprintf(strInfo, "%d.%d.%d.%d", 
				(unsigned char) addrIncoming.sa_data[2], 
				(unsigned char) addrIncoming.sa_data[3], 
				(unsigned char) addrIncoming.sa_data[4], 
				(unsigned char) addrIncoming.sa_data[5]);


			printf("[+] Accepted connection from: %s\n", strInfo);

			// Handle SSL errors.
			if(r <= 0)
			{
				int retError = SSL_get_error(ssl, r);

				if(r < 0)
				{
					printf("[-] ERROR: Fatal SSL accept error: %d\n", retError);
				}
				else
				{
					printf("[-] ERROR: SSL accept error - The TLS/SSL handshake was not successful but was shut down controlled: %d\n", retError);
				}

				if(SSL_ERROR_SSL)
				{
					char buf[2048];
					int err = ERR_get_error();
					ERR_error_string(err, buf);
					printf("SSL Error: %d - %s\n", err, buf);
				}
			}
			else
			{
				// Now we push the HTML-page.
				pushHTML(ssl, s, strInfo, bits, strShortname, strProtocol, testInfo);

				// Cleanup 
				int rCode = SSL_shutdown(ssl);

				if(!rCode)
				{
					shutdown(s, 1);
					rCode=SSL_shutdown(ssl);
				}
				  
				SSL_free(ssl);
				closesocket(s);
			}
      }
    }
}
